module fscounter(firclock, lrclock, firreset, fsgt90kHz, fsgt100kHz, fsgt190kHz);
input firclock, lrclock, firreset;
output reg fsgt90kHz;
output reg fsgt100kHz;
output reg fsgt190kHz;
reg [15:0] counter;
reg lrclocksynch1, lrclocksynch2, lrclockold;
always @ (posedge firclock) // firclock: 216 MHz = 8 * 27 MHz
begin
	lrclocksynch1 <= lrclock;
	lrclocksynch2 <= lrclocksynch1;
	lrclockold <= lrclocksynch2;
	if (firreset)
	begin
		fsgt90kHz <= 1'b1;
		fsgt100kHz <= 1'b1;
		fsgt190kHz <= 1'b0;
		// Het filter dat gebruikt wordt als 100 kHz < fs < 190 kHz werkt
		// altijd, dus neem 100 kHz < fs < 190 kHz maar als resetwaarde.
		// The filter that is used when 100 kHz < fs < 190 kHz always
		// works, so let's take 100 kHz < fs < 190 kHz as reset value.
		counter <= 16'b0;
	end
	else
	begin
		if (lrclocksynch2 && (!lrclockold))
		begin
			if (fsgt90kHz)
				fsgt90kHz <= (counter < 16'd2400); // 90 kHz
			else
				fsgt90kHz <= (counter < 16'd2374); // 90,98567818 kHz
			if (fsgt100kHz)
				fsgt100kHz <= (counter < 16'd2182); // 98,991750687 kHz
			else
				fsgt100kHz <= (counter < 16'd2160); // 100 kHz
			if (fsgt190kHz)
				fsgt190kHz <= (counter < 16'd1137); // 189,973614776 kHz
			else
				fsgt190kHz <= (counter < 16'd1131); // 190,98143236 kHz
			counter <= 16'b0;
		end
		else
			counter <= (counter + 16'b1) % 65536;
	end
end
endmodule
