module i2c(sda_in, sda_out, scl, sdclock, sdreset, address, read, readbyte, write, writebyte, ready);
// Primitieve I2C-interface die alleen maar een adres van een 24LC512 EEPROM
// kan lezen of schrijven en die geen gebruik maakt van de acknowledge-pulsen
// van de EEPROM om te zien of die eigenlijk wel klaar is met de laatste
// schrijfopdracht. Het gaat dus fout als je binnen 5 ms na een
// schrijfopdracht weer toegang tot de EEPROM probeert te krijgen.
// Primitive I2C interface that can only write to or read from an address of a 
// 24LC512 EEPROM and that doesn't use the acknowledge pulses to check whether
// the EEPROM has finished its last write command. Hence, things go wrong when
// the EEPROM is accessed within 5 ms after its last write command.
input sda_in, write, read, sdclock, sdreset;
input [15:0] address;
input [7:0] writebyte;
output sda_out, scl, ready;
output [7:0] readbyte;
reg sda_out_int, scl_int;
reg [7:0] readbyte_int;
reg [27:0] shifta;
reg [19:0] shiftro;
reg [9:0] shiftri;
reg [9:0] shiftw;
reg [7:0] counter1;
reg [7:0] counter;
reg count0, count111, count112, count149, count189;
reg pulse, shiftpulse, readshiftpulse;
reg active, active111, readnotwrite;
reg sf1, sf2, sf3;
reg [1:0] sum;
parameter deviceaddress = 7'b1010000; // 24LC512 met A2 = A1 = A0 = 0
// Rust: SDA = 1 en SCL = 1
always @ (posedge sdclock)
begin
	scl_int <= ~counter[1];
	if (sdreset)
		counter1 <= 8'b00000000;
	else
	begin
		if (active)
			counter1 <= (counter1 + 8'b00000001) % 256;
		else
			counter1 <= counter1;
	end
	if (sdreset || (count149 && (~readnotwrite)) || count189)
	begin
		active <= 1'b0;
		active111 <= 1'b0;
		counter <= 8'b00000000;
		readnotwrite <= 1'b1;
	end
	else
	begin
		active111 <= active111 || (active && count111);
		if (pulse) counter <= (counter + 8'b00000001) % 256;
		if (read)
		begin
			active <= 1'b1;
			readnotwrite <= 1'b1;
		end
		else
		begin
			if (write)
			begin
				active <= 1'b1;
				readnotwrite <= 1'b0;
			end
		end
	end
	if (active)
	begin
		if (shiftpulse && (~active111)) shifta[27:0] <= {shifta[26:0], 1'b0};
		if (shiftpulse && active111 && readnotwrite) shiftro[19:0] <= {shiftro[18:0], 1'b0};
		if (shiftpulse && active111 && (~readnotwrite)) shiftw[9:0] <= {shiftw[8:0], 1'b0};
		if (readshiftpulse && readnotwrite && active111) shiftri[9:0] <= {shiftri[8:0], sum[1]};
	end
	else
	begin // bits klaarzetten in de schuifregisters
	// prepare by parallel loading the bits into the shift registers
		shifta <= {1'b0, deviceaddress[6:0], 1'b0, 1'b1, address[15:8], 1'b1, address[7:0], 1'b1};
		shiftro <= {1'b0, deviceaddress[6:0], 1'b1, 1'b1, 9'b111111111, 1'b0};
		shiftw <= {writebyte[7:0], 1'b1, 1'b0};
		// Hiermee is het hoogste bit van shifta en van shiftro meteen
		// nul geworden, zoals noodzakelijk is om de start en herhaalde start
		// te maken.
		// This makes the MSBs of shifta and shiftro zero, as required
		// to make the start and repeated start signals.
	end
	if (count189) readbyte_int <= shiftri[9:2];
	sf1 <= sda_in;
	sf2 <= sf1;
	sf3 <= sf2;
end
always @ *
begin
	pulse = counter1[7] && counter1[6] && counter1[5] && counter1[4] && counter1[3] && counter1[2] && counter1[1] && counter1[0];
	readshiftpulse = pulse && (~counter[0]) && (~counter[1]);
	shiftpulse = pulse && (~counter[0]) && counter[1];
	if (counter == 8'b00000000)
		count0 = 1;
	else
		count0 = 0;
	if (counter == 8'b01101111)
		count111 = 1;
	else
		count111 = 0;
	if (counter == 8'b01110000)
		count112 = 1;
	else
		count112 = 0;
	if (counter == 8'b10010101)
		count149 = 1;
	else
		count149 = 0;
	if (counter == 8'b10111101)
		count189 = 1;
	else
		count189 = 0;
	// Omschakelen tussen schuifregisters en toevoegen start- en
	// stopcodes:
	// Switching between shift registers and adding start and stop codes:
	sda_out_int = (readnotwrite && (count111 || count112)) || count0 || (shifta[27] && active && (~active111)) || (shiftro[19] && readnotwrite && active111) || (shiftw[9] && active111 && (~readnotwrite)); 
	sum = sf1 + sf2 + sf3; // bit 1 van sum = meerderheid van sf1...sf3
	// bit 1 of sum = majority of sf1...sf3
end
assign sda_out = sda_out_int || sdreset; // voorkomt onterecht schrijven van de EEPROM als de boel nog niet synchroon gereset is
// prevents unintended EEPROM writing when everything isn't synchronously reset
// yet
assign scl = scl_int || sdreset; // voorkomt onterecht schrijven van de EEPROM als de boel nog niet synchroon gereset is
// prevents unintended EEPROM writing when everything isn't synchronously reset
// yet
assign readbyte = readbyte_int;
assign ready = ~active;
endmodule
