module integrator(x, clock, reset, y, clip, max);
// Integrator met vertraging, dus 1/(z-1),
// met tot +/-max begrensde toestandsvariabele,
// nauwkeurigheid toestandsvariabelebegrenzer beperkt.
// Integrator with delay, that is 1/(z-1),
// state variable limited to +/- max, limited accuracy
// of state variable limiting.
	parameter wordlength = 35;
	parameter limitlength = 10;
	input [limitlength-1:0] max;
	input signed [wordlength-1:0] x;
	input clock, reset;
	output signed [wordlength-1:0] y;
	reg signed [wordlength-1:0] y_int;
	output clip;
	reg clip_int;
	reg signed [wordlength:0] z;
	reg [limitlength:0] zabs;
	reg hulp;
	reg negatief;
	integer Piet;
    	always @ (posedge clock)
    	begin
		if (reset)
		begin
			y_int <= 0;
		end
		else
		begin
			if (clip_int)
			begin
				//y_int wordt exclusieve of van negatief en
				//max, dus max of -max-1.
				//y_int becomes the exclusive or of negatief
				//and max, that is, max or -max-1.
				if (negatief)
				begin
					y_int[wordlength-1] <= 1;
					y_int[wordlength-2:wordlength-limitlength-1] <= ~max;
					for (Piet=0; Piet<wordlength-limitlength-1; Piet=Piet+1)
						y_int[Piet] <= 1;
				end
				else
				begin
					y_int[wordlength-1] <= 0;
					y_int[wordlength-2:wordlength-limitlength-1] <= max;
					for (Piet=0; Piet<wordlength-limitlength-1; Piet=Piet+1)
						y_int[Piet] <= 0;
				end
			end
			else
				y_int[wordlength-1:0] <= {z[wordlength],z[wordlength-2:0]};
		end
	end
	always @ (*)
	begin
		z = y_int + x;
		//Als z[wordlength] ongelijk is aan z[wordlength-1], dan is
		//het resultaat in absolute waarde te groot voor wordlength
		//bits. In de volgende stap wordt het MSB van zabs dan "1".
		//Aangezien zabs een bit meer heeft dan max, is dan
		//automatisch zabs > max.
		//When z[wordlength] is unequal to z[wordlength-1], the result
		//is too large in absolute value to fit in wordlength bits. In
		//the next step, the MSB of zabs then becomes "1". As zabs is
		//one bit longer than max, this implies zabs > max.
		if (z[wordlength])
		       zabs[limitlength:0] = ~z[wordlength-1:wordlength-limitlength-1];
       		else
		       zabs[limitlength:0] = z[wordlength-1:wordlength-limitlength-1];
	        if (zabs > max)
			clip_int = 1;
		else
			clip_int = 0;
		negatief = z[wordlength];
	end
	assign y = y_int;
	assign clip = clip_int;
endmodule
