`default_nettype none
module knoplamp(sdclock, sdreset, modeswitch, filtswitch, bron, verrassingsknop,
	schaalknop, surprise, mod0, dsd, fsgt90kHz, fsgt100kHz, fsgt190kHz,
	qcliplampen, lampenknop, newsurpriserequest, relayoffkl, m, scale,
	filtsel, anacorr, sourcesel, halfbandlamp, apodizinglamp, steeplamp,
	chaoslamp, pwm4lamp, pwm8lamp, cliplamp, sclipl, sclipr, iclipl,
	iclipr, qclipl, qclipr, fclipinterpol, fclipprefilter);
input wire sdclock, sdreset;
// Goeie reset of heb je een latere nodig?
// Correct reset or do you need a later one?
input wire [3:0] modeswitch;
input wire [3:0] filtswitch;
input wire [3:0] bron;
input wire verrassingsknop;
input wire [2:0] schaalknop;
// pullup op alle schakelaars
// pull-up on all switches
PULLUP pums3(modeswitch[3]);
PULLUP pums2(modeswitch[2]);
PULLUP pums1(modeswitch[1]);
PULLUP pums0(modeswitch[0]);
PULLUP pufs3(filtswitch[3]);
PULLUP pufs2(filtswitch[2]);
PULLUP pufs1(filtswitch[1]);
PULLUP pufs0(filtswitch[0]);
PULLUP pubr3(bron[3]);
PULLUP pubr2(bron[2]);
PULLUP pubr1(bron[1]);
PULLUP pubr0(bron[0]);
PULLUP puvk(verrassingsknop);
PULLUP pusk2(schaalknop[2]);
PULLUP pusk1(schaalknop[1]);
PULLUP pusk0(schaalknop[0]);
input wire [3:0] surprise; // verrassingswaarde van sequencer.v, surprise value from sequencer.v
input wire mod0, dsd, fsgt90kHz, fsgt100kHz, fsgt190kHz; // sigma-delta in modus 0 (stilte) dwingen, DoP-detectoruitgang, fs groter dan 90, 100 of 190 kHz
// signals to force the sigma-delta into mode 0 (silence), DoP detector output, fs greater than 90, 100 or 190 kHz

input wire qcliplampen; // DIL-schakelaartje waarmee je kiest of vastlopen kwantisator tot oplichten oversturingslamp leidt
// DIL-switch that sets whether quantizer clipping activates the clipping lamp

input wire lampenknop; // DIL-schakelaartjes waarmee je kiest of oversturingslamp ook werkt in de verrassingsstanden
// DIL-switch that sets whether the clipping lamp works in the surprise modes
PULLUP puqcle(qcliplampen);
PULLUP pulek(lampenknop);

input wire sclipl, sclipr, iclipl, iclipr, qclipl, qclipr, fclipinterpol, fclipprefilter;
output reg newsurpriserequest; // verzoek aan sequencer om een nieuwe verrassingswaarde te trekken
// requests the sequencer to generate a new surprise value

output wire relayoffkl;
output reg [1:0] m;
output reg [1:0] scale;
output reg [2:0] filtsel;
output reg anacorr; // 1 als het interpolatiefilter moet corrigeren voor de afval van het analoge filter
// 1 when the interpolation filter has to correct the roll-off of the analogue filter

output reg [1:0] sourcesel;
output reg halfbandlamp, apodizinglamp, steeplamp, chaoslamp, pwm4lamp, pwm8lamp;
output wire cliplamp;
reg [12:0] knoppen;
reg [12:0] knoppentoestand;
reg [12:0] knoppentoestand2;
reg [19:0] ontdenderteller;
reg [19:0] relaisuitteller;
reg [2:0] schaalknop2;
reg verrassing1, verrassing2;
reg [1:0] fls2;
reg [26:0] scliplampcounter;
reg [24:0] icliplampcounter;
reg [8:0] qcliplampcounter;
reg lampen, lampff;
reg verrassingsknop2oud;
wire [3:0] modeswitch2;
wire [3:0] filtswitch2;
wire [3:0] bron2;
wire verrassingsknop2;
wire iclipsupp;


// ontdenderaar en uitschakelen uitgangsrelais en oversturingslamp, schaalknop
// gaat hier niet doorheen	
//
// debouncer and switching off the output relay and the clipping lamp,
// schaalknop is not processed here
//
always @ (posedge sdclock)
begin
	knoppen <= {modeswitch[3:0], filtswitch[3:0], bron[3:0], verrassingsknop};
	if (sdreset) // ?
	begin
		ontdenderteller <= 20'b0;
		relaisuitteller <= 20'hFFFFF;
		knoppentoestand <= knoppen;
		knoppentoestand2 <= knoppen;
	end
	else
	begin
		if (knoppen != knoppentoestand)
		begin
			if (ontdenderteller[19])
			begin
				knoppentoestand <= knoppen;
				if (!((!knoppen[0]) && knoppentoestand[0])) relaisuitteller <= 20'hFFFFF;
// Het geluid en het oversturingslichtje gaan uit als relaisuitteller
// ongelijk aan 0 is.
// De voorwaarde betekent dat de verrassingsknop niet net
// ingedrukt is, bij indrukken van de verrassingsknop hoeft het geluid
// niet uit. knoppen[0] wordt 0 als de verrassingsknop wordt ingedrukt.
//
// The sound and the clipping lamp are switched off when relaisuitteller
// is not equal to zero.
// The condition means that the sound is not switched off when the surprise
// button is depressed (knoppen[0] = 0).
			end
			else
				ontdenderteller <= ontdenderteller + 20'b1;
		end
		else
		begin
			ontdenderteller <= 20'b0;
			if (relaisuitteller != 20'b0) relaisuitteller <= relaisuitteller - 20'b1;
		end
		if (relaisuitteller[19:17] == 3'b011) knoppentoestand2[12:1] <= knoppentoestand[12:1];
		knoppentoestand2[0] <= knoppentoestand[0];
	end
end
assign modeswitch2[3:0] = knoppentoestand2[12:9];
assign filtswitch2[3:0] = knoppentoestand2[8:5];
assign bron2[3:0] = knoppentoestand2[4:1];
assign verrassingsknop2 = knoppentoestand2[0];
assign iclipsupp = (relaisuitteller != 20'b0);
assign relayoffkl = iclipsupp;

// omschakelen modus sigma-delta
// switching the sigma-delta mode
always @ *
begin
	case ({mod0, modeswitch2[3:0]})
		5'b01110: // chaos
		begin
			m = 2'b01;
			verrassing1 = 1'b0;
		end
		5'b01101: // PBM4, PWM4
		begin
			m = 2'b10;
			verrassing1 = 1'b0;
		end
		5'b01011: // PBM8, PWM8
		begin
			m = 2'b11;
			verrassing1 = 1'b0;
		end
		5'b00111: // verrassing, surprise
		begin
			m = surprise[1:0];
			verrassing1 = 1'b1;
		end
		default:
		begin
			m = 2'b00; // stilte, silence
			verrassing1 = 1'b1; // oversturingslampje uit, clipping lamp forced off
		end
	endcase
end

// omschakelen prefilter, switching prefilter
//
// prefilter:		0XX: niet filteren / no filtering
// 			100: steil filter met tweevoudige interpolatie, fs <= 100 kHz / steep two-times-interpolating filter, fs <= 100 kHz
// 			101: DSD-filter
// 			110: glooiend filter dat tempert ofte apodiseert / apodizing filter
// 			111: transitiebandversmaller zonder interpolatie, overbodig
// 				voor fs >= 190,144 kHz omdat de Nyquist-frequentie dan
// 				in de stopband van de interpolatieketen (module interpol) ligt
// 			     transition band narrowing filter without interpolation,
// 			        not needed for fs >= 190.144 kHz because the
// 			        Nyquist frequency then lies in the stop band
// 			        of th interpolation chain (module interpol)
// filtswitch2[3:0]:	0111: halfband
// 			1011: glooiend / apodizing
// 			1101: steil / steep
// 			1110: verrassing / surprise
// 			alle andere (ongeldige) waarden: glooiend / any other (invalid) value: apodizing
// surprise[3:2]:	01: halfband
// 			10: glooiend / apodizing
// 			11: steil / steep
always @ *
begin
	if (dsd)
	begin
		filtsel = 3'b101;
		anacorr = 1'b1;
		verrassing2 = 1'b0; // Bij DSD gebruik je altijd het DSD-filter en kan er wat filtering betreft dus geen sprake zijn van enige verrassing.
				    // In DSD mode you always use the DSD filter, so it is no surprise what filter is used.
	end
	else
	begin
		if (filtswitch2[3:0] == 4'b1110)
			verrassing2 = 1'b1;
		else
			verrassing2 = 1'b0;
		case (filtswitch2[3:0])
			4'b0111: fls2 = 2'b01;
			4'b1011: fls2 = 2'b10;
			4'b1101: fls2 = 2'b11;
			4'b1110: fls2 = surprise[3:2];
			default: fls2 = 2'b10;
		endcase
		if ((!fsgt90kHz) && (fls2 == 2'b10))
			anacorr = 1'b0; // Voor fs < 90 kHz piekt het
// glooiende filter meer dan het analoge filter afvalt -> geen correctie
// nodig.
//
// At fs < 90 kHz, the apodizing filter peaks more than the analogue filter
// rolls off, so there is no correction needed.
//
		else
			anacorr = 1'b1;
		case (fls2[1:0])
			2'b01: filtsel = 3'b000; // halfband
			2'b11: // steil / steep
			begin
				case ({fsgt190kHz, fsgt100kHz})
					2'b00: filtsel = 3'b100; // onder de 100 kHz -> kan interpoleren
								 // below 100 kHz -> interpolate
					2'b11: filtsel = 3'b000; // boven de 190 kHz -> interpolatieketen onderdrukt spectrale kopieen wel
								 // above 190 kHz -> interpolation chain will suppress spectral copies
					default: filtsel = 3'b111; // steil filter dat niet interpoleert
								   // steep non-interpolating filter
				endcase
			end
			default: filtsel = 3'b110; // glooiend / apodizing
		endcase
	end
end

// Oversturingslampaansturing, lamp brandt 16 klokpulsen (als hij zo
// snel aan gaat althans) bij qclip mits
// qcliplampen, ongeveer vijf seconden bij sclip of fclip (die horen niet
// voor te kunnen komen), ruim 1 seconde geleidelijk gedimd bij iclip.
// Met lampenknop kun je kiezen of de lamp ook moet werken in de
// verrassingsstand.
// Het lamp uitschakelen in de verrassingsstand moet zodanig gebeuren dat er
// bij het omschakelen van de verrassingsstand naar de gewone stand ook
// niet even een lichtje gaat branden! Anders kun je zien dat er een transitie
// heeft plaatsgevonden en dat de verrassingsstand dus kennelijk niet de stand
// was waar je 'm nu op gezet hebt.
//
// Clipping lamp control, lamp glows 16 cycles (if it ignites that
// quickly) at qclip when qcliplampen,
// about five seconds when sclip or fclip (which should be impossible), a bit
// more than 1 second gradually dimmed when iclip. With lampenknop you can
// choose whether it should work in surprise mode.
// Switching off the lamp in surprise mode must be done such that it does not
// light up when you switch from surprise mode to a normal mode!
//
always @ (posedge sdclock)
begin
	if (sdreset)
	begin
		scliplampcounter <= 27'b0;
		icliplampcounter <= 25'b0;
		qcliplampcounter <= 9'b0;
	end
	else
	begin
		if ((sclipl || sclipr || fclipinterpol || fclipprefilter) && lampen)
			scliplampcounter <= 27'h7FFFFFF; // 2^27 - 1
		else
			if (scliplampcounter != 27'b0) scliplampcounter <= scliplampcounter - 27'b1;
		if ((iclipl || iclipr) && (!iclipsupp) && lampen)
			icliplampcounter <= 25'h1FFFFFF; // 2^25 - 1
		else
			if (icliplampcounter != 25'b0) icliplampcounter <= icliplampcounter - 25'b1;
		if ((qclipl || qclipr) && qcliplampen && lampen)
			qcliplampcounter <= 9'h010; // 16
		else
			if (qcliplampcounter != 9'b0) qcliplampcounter <= qcliplampcounter - 9'b1;
		lampff <= (scliplampcounter != 27'h0000000) || (icliplampcounter[7:0] < icliplampcounter [24:17]) || (qcliplampcounter != 9'h000);
	end
end
always @ * lampen = (!(verrassing1 || verrassing2)) || lampenknop;
assign cliplamp = lampff;

// schaalfactor instellen, gaat NIET door ontdenderblok!
// setting the scaling factor, NOT debounced!
always @ (posedge sdclock)
begin
	schaalknop2 <= schaalknop;
	case (schaalknop2[2:0])
		3'b110: scale <= 2'b00; // zacht, gegarandeerd niet overstuurd / soft
		3'b101: scale <= 2'b01; // middel / medium
		3'b011: scale <= 2'b10; // hard / loud mode
					// sigmadelta.v kent ook een stand 11,
					// maar die slaat eigenlijk nergens op
					// sigmadelta.v also has a setting 11,
					// but that one makes no sense
		default: scale <= scale;
	endcase
// Instellen bron gaat wel met een ontdenderde bronkeuze.
// Als de schakelaar even geen contact maakt, bijvoorbeeld omdat je heel
// langzaam aan de knop draait, houdt hij de oude toestand vast.
//
// Source selection, this is debounced.
	case (bron2[3:0])
		4'b1110: sourcesel <= 2'b00;
		4'b1101: sourcesel <= 2'b01;
		4'b1011: sourcesel <= 2'b10;
		4'b0111: sourcesel <= 2'b11;
		default: sourcesel <= sourcesel;
	endcase
end

// Uitlezen verrassing wanneer de verrassingsknop ingedrukt gehouden wordt.
// Reading off the surprise when the surprise button is depressed.
// verrassingsknop2 = 0: knop ingedrukt / button depressed
// verrassingsknop2 = 1: knop niet ingedrukt / button not depressed
//
always @ *
begin
	halfbandlamp = (!surprise[3]) && surprise[2] && (!verrassingsknop2);
	apodizinglamp = surprise[3] && (!surprise[2]) && (!verrassingsknop2);
	steeplamp = surprise[3] && surprise[2] && (!verrassingsknop2);
	chaoslamp = (!surprise[1]) && surprise[0] && (!verrassingsknop2);
	pwm4lamp = surprise[1] && (!surprise[0]) && (!verrassingsknop2);
	pwm8lamp = surprise[1] && surprise[0] && (!verrassingsknop2);
end

// aanvragen nieuwe verrassingswaarde
// requesting new surprise value
always @ (posedge sdclock)
begin
	if (sdreset) // Goede reset? Correct reset?
	begin
		newsurpriserequest <= 1'b0;
		verrassingsknop2oud <= 1'b1; // los, want 0 is ingedrukt
					     // button not pressed, as 0 means pressed
	end
	else
	begin
		verrassingsknop2oud <= verrassingsknop2;
		newsurpriserequest <= verrassingsknop2 && (!verrassingsknop2oud);
	end
end
endmodule
`default_nettype wire
