`default_nettype none
module prefilter(reset, bck, lrck, data, firclock, sel, dither, bckout, lrckout, dataout);
// sel[2:0] = 0XX: niet filteren, alleen verzwakken
// sel[2:0] = 100: twee keer interpolatie
// sel[2:0] = 101: DSD-filter
// sel[2:0] = 110: temperen ofte apodiseren
// sel[2:0] = 111: transitieband versmallen zonder interpolatie
// 
// sel[2:0] = 0XX: don't filter, just attenuate
// sel[2:0] = 100: two times interpolation
// sel[2:0] = 101: DSD filter
// sel[2:0] = 110: apodize
// sel[2:0] = 111: narrow the transition band without interpolation
//
// sel[2:0] moet synchroon zijn met firclock, anders gaat het opwekken van
// sclr5...sclr7 fout.
// sel[2:0] must be synchronized to firclock, otherwise generating
// sclr5...sclr7 goes wrong.
input wire reset, bck, lrck, data, firclock;
input wire [2:0] sel;
input wire signed [64:0] dither;
output wire bckout, lrckout, dataout;
reg [64:0] dithers;
reg signed [31:0] din;
reg bcks, lrckds, datas, lr1, lr2, lr3, lr4, lr5, b1, b2, d1, d2, d3;
reg bckup, lrup, lrdown, sclr4, sclr5, sclr6, sclr7, rd_en, zus;
wire nd4, nd6, nd7;
reg hulp, hulp1, hulp2, hulp3, hulp4;
reg bckor128fs;
reg notbckor128fs;
reg nd5;
reg aap, noot, mies, wim;
reg [5:0] bitcount;
reg dsde;
wire signed [64:0] dout4;
wire signed [34:0] dout5l;
wire signed [34:0] dout5r;
wire signed [64:0] dout6;
wire signed [65:0] dout7;
reg [67:0] inmux0;
reg [67:0] inmux4;
reg [67:0] inmux5;
reg [67:0] inmux6;
reg [67:0] inmux7;
reg [67:0] muxout;
wire [66:0] fifodout;
reg [65:0] fifodoutd1;
reg [23:0] fifodoutdit;
wire rdy4, rdy5l, rdy5r, rdy6, rdy7, chan_out4, chan_out6, chan_out7;
reg rdy5ld;
wire chan_in4, chan_in6, chan_in7;
wire rfd4, rfd5l, rfd5r, rfd6, rfd7;
reg [2:0] sel_old;
reg [4:0] readcounter;
reg fifowaitff, resetfiltsw;
wire resetfiltsw1, resetfiltsw2;
reg dataoutt;
reg tiehigh, pisopl;
reg [30:0] piso;
wire full, empty, prog_empty;
always @ (posedge bck)
begin
	lr1 <= lrck;
	lr2 <= lr1;
	d1 <= data;
end
always @ (negedge bck)
begin
	lr3 <= lr2;
	d2 <= d1;
	din[31:0] <= {din[30:0], d2};
end
always @ (posedge firclock)
begin
	b1 <= bck;
	bcks <= b1;
	lr4 <= lr3;
	lrckds <= lr4;
	d3 <= d2;
	datas <= d3;
	b2 <= bcks;
	lr5 <= lrckds;
	aap <= din[31];
	noot <= aap;
end
always @ *
begin
	bckup = bcks & (~b2);
	lrup = lrckds & (~lr5);
	lrdown = lr5 & (~lrckds);
	mies = ~noot;
	wim = ~datas;
end
always @ (posedge firclock)
begin
	if (lrup)
		bitcount <= 6'b000000;
	else
	begin
		if (bckup)
			bitcount <= bitcount + 6'b000001;
		else
			bitcount <= bitcount;
	end
end
always @ *
begin
	case (bitcount[5:3])
		3'b001: dsde = 1;
		3'b010: dsde = 1;
		default: dsde = 0;
	endcase
	// dsde geeft bij DoP aan welke bits de echte DSD-data bevatten.
	// Detecteren of er sprake is van DoP gebeurt niet in deze module.
	//
	// In case of DoP, dsde indicates which bits contain the actual DSD
	// data. This module does not detect whether the signal is DoP at all.
	//
	nd5 = dsde & bckup;
end
firdsd firdsdl (
	.sclr(sclr5), // input sclr
	.clk(firclock), // input clk
	.nd(nd5), // input nd
	.rfd(rfd5l), // output rfd
	.rdy(rdy5l), // output rdy
	.din({mies, 1'b1}), // input [1 : 0] din
	.dout(dout5l)); // output [34 : 0] dout
firdsd firdsdr (
	.sclr(sclr5), // input sclr
	.clk(firclock), // input clk
	.nd(nd5), // input nd
	.rfd(rfd5r), // output rfd
	.rdy(rdy5r), // output rdy
	.din({wim, 1'b1}), // input [1 : 0] din
	.dout(dout5r)); // output [34 : 0] dout
// DSD-filters, 2 bit in, 35 bit uit, 14 keer decimerend,
// Wilkinson N = 256, p = 3, tau = 0, stopband vanaf 0,03239285714 fs.
//
assign nd4 = (chan_in4 ? lrdown : lrup) && (!sclr4);
fir384interpol2 fir384interpol2lr (
	.sclr(sclr4), // input sclr
	.clk(firclock), // input clk
	.nd(nd4), // input nd
	.rfd(rfd4), // output rfd
	.rdy(rdy4), // output rdy
	.chan_in(chan_in4), // output [0 : 0] chan_in
	.chan_out(chan_out4), // output [0 : 0] chan_out
	.din(din), // input [31 : 0] din
	.dout(dout4)); // output [64 : 0] dout
// twee keer interpolerend filter, interpoleert
// met een macht van twee en hoeft dus niet geschaald te worden.
//
// two times interpolating filter, interpolates with a power of two and,
// hence, need not be scaled.
//
assign nd6 = (chan_in6 ? lrdown : lrup) && (!sclr6);
fir40 fir40lr (
	.sclr(sclr6), // input sclr
	.clk(firclock), // input clk
	.nd(nd6), // input nd
	.rfd(rfd6), // output rfd
	.rdy(rdy6), // output rdy
	.chan_in(chan_in6), // output [0 : 0] chan_in
	.chan_out(chan_out6), // output [0 : 0] chan_out
	.din(din), // input [31 : 0] din
	.dout(dout6)); // output [64 : 0] dout
// temper- ofte apodiseringsfilter
// Uitgangssignaal moet een bitje verschoven worden ten opzichte van de
// ruigere filters vanwege het beperkte doorschot.
//
// apodizing filter, output signal must be shifted a bit compared to the
// steeper filters because the apodizing filter has much less overshoot.
//
assign nd7 = (chan_in7 ? lrdown : lrup) && (!sclr7);
fir512 fir512lr (
	.sclr(sclr7), // input sclr
	.clk(firclock), // input clk
	.nd(nd7), // input nd
	.rfd(rfd7), // output rfd
	.rdy(rdy7), // output rdy
	.chan_in(chan_in7), // output [0 : 0] chan_in
	.chan_out(chan_out7), // output [0 : 0] chan_out
	.din(din), // input [31 : 0] din
	.dout(dout7)); // output [65 : 0] dout
// steil niet-interpolerend filter
// steep non-interpolating filter
always @ (posedge firclock) rdy5ld <= rdy5l;
always @ *
begin
	inmux0 = {(lrup|lrdown), (~lrckds), din[31], din[31], din[31:0], 32'b0};
	inmux4 = {rdy4, chan_out4, dout4[64:0], 1'b0};
	zus = rdy5l|rdy5ld;
	if (rdy5l)
		inmux5 = {zus, (~rdy5l), dout5l[34:0], 31'b0};
	else
		inmux5 = {zus, (~rdy5l), dout5r[34:0], 31'b0};
	inmux6 = {rdy6, chan_out6, dout6[64], dout6[64:0]};
	inmux7 = {rdy7, chan_out7, dout7[65:0]};
	case (sel[2:0])
		3'b100: muxout = inmux4;
		3'b101: muxout = inmux5;
		3'b110: muxout = inmux6;
		3'b111: muxout = inmux7;
		default: muxout = inmux0;
	endcase
end
fifo16_67 fifo (
  .rst(resetfiltsw), // input rst
  .wr_clk(firclock), // input wr_clk
  .rd_clk(bckor128fs), // input rd_clk
  .din(muxout[66:0]), // input [66 : 0] din
  .wr_en(muxout[67]), // input wr_en
  .rd_en(rd_en), // input rd_en
  .dout(fifodout), // output [66 : 0] dout
  .full(full), // output full
  .empty(empty), // output empty
  .prog_empty(prog_empty) // output prog_empty
);
always @ *
begin
	fifodoutd1[65:0] = $signed(fifodout[65:0]) + $signed(dithers[64:22]);
	fifodoutdit = $signed(fifodoutd1[65:42]);
end
always @ (posedge firclock)
begin
	case (sel[2:0]) // case statement used to be in an always @ * block, put here as pipelining
		3'b100:
		begin
			sclr4 = reset;
			sclr5 = 1'b1;
			sclr6 = 1'b1;
			sclr7 = 1'b1;
		end
		3'b101:
		begin
			sclr4 = 1'b1;
			sclr5 = reset;
			sclr6 = 1'b1;
			sclr7 = 1'b1;
		end
		3'b110:
		begin
			sclr4 = 1'b1;
			sclr5 = 1'b1;
			sclr6 = reset;
			sclr7 = 1'b1;
		end
		3'b111:
		begin
			sclr4 = 1'b1;
			sclr5 = 1'b1;
			sclr6 = 1'b1;
			sclr7 = reset;
		end
		default:
		begin
			sclr4 = 1'b1;
			sclr5 = 1'b1;
			sclr6 = 1'b1;
			sclr7 = 1'b1;
		end
	endcase
	sel_old[2:0] = sel[2:0];
end
always @ *
begin
	resetfiltsw = reset|(sel_old[2]^sel[2])|(sel_old[1]^sel[1])|(sel_old[0]^sel[0]);
	tiehigh = 1'b1;
end
FDPE ffasynchset1(resetfiltsw1, bckor128fs, tiehigh, resetfiltsw, resetfiltsw); //Q, C, CE, PRE, D
FDPE ffasynchset2(resetfiltsw2, bckor128fs, tiehigh, resetfiltsw, resetfiltsw1); //Q, C, CE, PRE, D
always @ (posedge bckor128fs)
begin
	fifowaitff <= (~resetfiltsw2) & ((~prog_empty)|fifowaitff);
	if (resetfiltsw2|rd_en)
		readcounter <= 5'b00000;
	else
	begin
		if (fifowaitff)
			readcounter <= readcounter + 5'b00001;
		else
			readcounter <= readcounter;
	end
end
always @ * rd_en = readcounter[4] & readcounter[3] & readcounter[1] & readcounter[0] & (readcounter[2]|(sel[2] & (~sel[1]) & sel[0]));
always @ (posedge bckor128fs)
begin
	dithers <= dither; //pijpleidingerij, pipelining
	pisopl <= rd_en;
	if (pisopl)
	begin
		dataoutt <= fifodoutdit[23];
		piso[30:0] <= {fifodoutdit[22:0], 8'b00000000};
	end
	else
	begin
		dataoutt <= piso[30];
		piso[30:0] <= {piso[29:0], 1'b0};
	end
end

always @ (posedge bck) hulp <= (~sel[2]) | sel[1] | sel[0];
// hulp = 1 als er geen dubbele klok nodig is
// hulp = 1 when there is no doubled clock needed

reg hulp5, hulp6, hulp7, hulp8, hulp9, hulp10, hulpa, hulpb;
always @ (posedge firclock)
begin
	hulpa <= hulp;
	hulpb <= hulpa;
	hulp1 <= bck;
	hulp2 <= hulp1;
	hulp3 <= hulp2 | hulpb;
	hulp4 <= hulp3;
	hulp5 <= hulp4;
	hulp6 <= hulp5;
	hulp7 <= hulp6;
	hulp8 <= hulp7;
	hulp9 <= hulp8;
	hulp10 <= hulp9;
end

always @ (posedge firclock) notbckor128fs <= hulp2 ^ hulp10;
always @ * bckor128fs = ~notbckor128fs;

assign bckout = notbckor128fs;
assign lrckout = fifodout[66];
assign dataout = dataoutt;
endmodule
`default_nettype wire
