module quantizer(x, m, rnd1, rnd2, y, clip);
parameter wordlength = 35;
input signed [wordlength-1:0] x;
input signed [wordlength-4:0] rnd1; //uniform verdeeld, uniform distribution
input signed [wordlength-4:0] rnd2; //uniform verdeeld, uniform distribution
input [1:0] m; //00: uit (off), 01: twee niveaus (two levels), 10: vijf niveaus (five levels), 11: negen niveaus (nine levels)
output signed [3:0] y; // Ook bij een eenbitter is 0 dus positief. Also in the single bit case, 0 means positive.
output clip; 
// Half uitgestuurd komt overeen met volle schaal, verder -> clip.
// Driven halfway means full scale, further -> clip.
reg signed [wordlength-3:0] dither;
reg signed [wordlength-3:0] offcorr;
reg signed [4:0] bla;
reg signed [3:0] y_int;
reg clip_int;
reg signed [wordlength:0] sum;
always @ *
begin
case(m)
	2'b00:
	begin
		dither = 0;
		offcorr = 0;
	end
	2'b01:
	begin
		dither = 0;  // 1 bit -> geen ruimte voor dither (no space for dither)
		offcorr = 0; // beslissingsgrens bij 0 (decision threshold at 0)
	end
	2'b10:
	begin
		dither = rnd1 + rnd2; //vijfwaardig -> volledige dither (five levels -> full dither)
		offcorr = {1'b0, 1'b1, {(wordlength-4) {1'b0}}};
	end
	2'b11:
	begin
		dither = (rnd1>>>1) + (rnd2>>>1); //negenwaardig -> dither bitje zachter (nine levels -> dither one bit softer)
		offcorr = {2'b0, 1'b1, {(wordlength-5) {1'b0}}};
	end
endcase
sum = x+dither+offcorr;
bla[4:0] = sum[wordlength:wordlength-4];
if ((bla[4:2] == 3'b000)|(bla[4:2] == 3'b111)|((m == 2'b10)&(bla[4:1] == 4'b0010))|((m == 2'b11)&(bla[4:0] == 5'b00100)))
// De wat ingewikkelde uitdrukking achter de if is nodig vanwege de asymmetrie
// in tweecomplementcode. Voor m = 2 zijn alle uitgangscodes van -2 tot en met
// +2 toegestaan en voor m = 3 alle uitgangscodes van -4 tot en met +4. De
// eerste twee termen dekken echter de +2 respectievelijk +4 niet af.
//
// This somewhat complicated expression after the if is needed because of the
// asymmetry in two's complement codes. For m = 2, all output codes from -2
// up to and including +2 are allowed and for m = 3, all output codes from
// -4 up to and including +4. The first two terms of the expression after the
// if do not cover +2 respectively +4.
//
begin
	clip_int = 0;
	case(m)
		2'b00: y_int[3:0] = 4'b0000;
		2'b01: y_int[3:0] = {bla[3], bla[3], bla[3], bla[3]};
		2'b10: y_int[3:0] = {bla[3], bla[3], bla[2], bla[1]};
		2'b11: y_int[3:0] = bla[3:0];
	endcase
end
else
begin
	clip_int = 1;
	if (bla[4] == 0)
	begin
		case (m)
			2'b00: y_int[3:0] = 4'b0000; 
			2'b01: y_int[3:0] = 4'b0000;
			2'b10: y_int[3:0] = 4'b0010;
			2'b11: y_int[3:0] = 4'b0100;
		endcase
	end
	else
	begin
		case (m)
			2'b00: y_int[3:0] = 4'b0000;
			2'b01: y_int[3:0] = 4'b1111;
			2'b10: y_int[3:0] = 4'b1110;
			2'b11: y_int[3:0] = 4'b1100;
		endcase
	end
end
end
assign y = y_int;
assign clip = clip_int;
endmodule
