module sequencer(clock24, reset, sdclock, sdreset, sourcesel, rnd,
	newsurpriserequest, dix4192cdout, src4392csn, dix4192csn, src4392cclk,
	dix4192cclk, srcdixrstn, src4392cdin, dix4192cdin, switch85A2,
	ignited85A2, seed, scl, sda_in, sda_out, relayoffkl, relay,
	knoplampreset, surprise);
// Inclusief aansturen van de EEPROM, I2C-interface komt hieronder te hangen.
// Deze module kijkt of de 27 MHz-klok ongeveer de juiste
// frequentie heeft en detecteert wanneer dat het geval is.
// De tijd tot de 27 MHz er is kun je ook voor het opwekken
// van het zaadje gebruiken, net als de starttijd der 85A2. 
//
// Includes controlling the EEPROM, the I2C interface is instantiated
// in this module.
// This module checks whether the 27 MHz clock has about the right
// frequency by comparing it to the 24 MHz from the FPGA module.
// The 27 MHz start-up time can be used for the seeds of the
// pseudorandom number generators, just like the 85A2 ignition delay.

input clock24, sdclock, ignited85A2, sda_in, newsurpriserequest, dix4192cdout;
input relayoffkl;
input [1:0] sourcesel;
output reset, sdreset, src4392csn, src4392cclk, dix4192csn, dix4192cclk;
output srcdixrstn, src4392cdin, dix4192cdin, switch85A2, scl, sda_out, relay;
output knoplampreset;
output [31:0] seed;
output reg [3:0] surprise;
input [5:0] rnd; // Zowel voor trekken verrassingswaarde als voor verruisen updater.
// Used for generating surprise values and for making the DIX4192 update
// signals more noise-like.
reg [3:0] rndd;
reg spicsn_int, spicclk_int, src4392cdin_int, dix4192cdin_config;
reg switch85A2_int, spirstn_int;
reg knoplampreset_int;
reg relay_int;
reg [14:0] teller;
reg [2:0] tel3;
reg [7:0] tel8;
reg [10:0] tel11;
reg [7:0] reg8;
reg sdclkdiv8_1, sdclkdiv8_2, sdclkdiv8_3;
reg [26:0] delaycount27;
reg [15:0] seedcount1;
reg [15:0] seedcount2;
reg dixsrcreset, i2cready_old;
reg [2:0] vt; // toestandsvariabele verrassingstoestandmachine, state variable surprise state machine
reg [19:0] relaycounter;
reg waitfori2c;
wire endofstartup, reset0, reset1, reset2;
wire resetsynch1, resetsynch2, i2cready, sdreset1, sdreset2;
reg read, write, updaterreset;
wire dix4192cdin_ud, dix4192cclk_ud, dix4192csn_ud;
wire dixupdaterreset, dixupdaterreset1;
wire [7:0] readbyte;
wire [7:0] writebyte;
parameter [511:0] src4392spidata = 512'h0100390031004100000000000000081821CB02000000000000000000000000000000000000000000000000000000010400000000000000000000000000000000;
// Deze bits moeten van hoog naar laag erin geschoven worden!
// Adres 01, waarde 39: Bemonsteringsfrequentieomzetter aan, DIR uit, DIT uit,
// poort B aan, poort A aan, algemene enable aan, reset uit.
// adres 03, waarde 31: poort A in slaafstand, signaal van SRC eruit, geluid
// niet uit, 24 bits I2S
// adres 05, waarde 41: poort B in slaafstand, uitgang uit, 24 bits I2S
// adressen 0B en 0C, waarde 00: nu staat er niks aan, maar wil je nog wat
// met interrupties van de SRC?
// adres 0D, waarde 08: RXCLK op MCLK gezet, ingang 0 naar DIR. Maakt weinig
// uit, want we gebruiken de DIR toch niet.
// adres 0E, waarde 18: geluid uit de DIR uit wanneer de DIR ontgrendeld
// raakt, klok van PLL2 niet afdelen, RXCKO-uitgang uit, PLL2 loopt vrij
// als zij niet vergrendelt
// adressen 0F...11, waarden 21CB02: PLL1 op P = 2, J = 7, D = 2818, zoals
// aanbevolen met 27 MHz referentieklok
// adres 1B, waarde 00: GPO1 uit
// adres 1C, waarde 00: GPO2 uit
// adres 1D, waarde 00: GPO3 uit
// adres 1E, waarde 00: GPO4 uit
// adressen 1F...2C: alleen leesbaar, dus geschreven waarde slaat nergens op.
// adres 2D, waarde 01: SRC converteert data van poort B, gebruikt de MCLK als
// referentie, geluid niet uit, sterktes links en rechts apart instelbaar
// adres 2E, waarde 04: de-emfase uit, geen decimatiefiltering, FIFO
// van 64 woorden lang
// adres 2F, waarde 00: 24 bits woorden uit
// adressen 30 en 31, waarde 00: uitgangsverzwakkers op 0 dB
// Rest is onzin.
//
// Configuration data for the SRC4392, see data sheet for an explanation
//
parameter [511:0] dix4192spidata = 512'h0100320029080000000000000000081D21CB020000000000000000000605070D0000000000000000000000000000000000000000000000000000000000000000;
// Deze bits moeten van hoog naar laag erin geschoven worden!
// Adres 01, waarde 32: DIR aan, DIT uit, poort B uit, poort A aan,
// algemene enable aan, reset uit.
// adres 03, waarde 29: poort A in meesterstand, signaal van SPDIF-
// ontvanger eruit, geluid niet uit, 24 bits I2S
// adres 04, waarde 08: RXCKO gedeeld door 128 als LRCKA
// adres 0D, waarde 08: RXCLK op MCLK gezet, ingang 0 naar DIR
// adres 0E, waarde 1D: geluid uit wanneer de DIR ontgrendeld raakt, klok van
// PLL2 door vier delen (CD-waarde, wordt later bijgewerkt indien nodig),
// RXCKO-uitgang aan, PLL2 loopt vrij als zij niet vergrendelt
// adressen 0F...11, waarden 21CB02: PLL1 op P = 2, J = 7, D = 2818, zoals
// aanbevolen met 27 MHz referentieklok
// adres 1B, waarde 06: GPO1 hoog als er geen audio-data ontvangen worden
// adres 1C, waarde 05: GPO2 laag als er pre-emfase gebruikt is
// adres 1D, waarde 07: GPO3 hoog als er ongeldige data ontvangen worden
// adres 1E, waarde 0D: GPO4 hoog bij een pariteitsfout
// adressen 1F...2C: alleen leesbaar, dus geschreven waarde slaat nergens op.
// adres 2D en verder: bestaat niet bij de DIX4192, dus allemaal onzin.
//
// Configuration data for the DIX4192, see data sheet for an explanation
//

// Opwekken algehele reset
// Generating the overall reset

STARTUP_SPARTAN6 startding(.EOS(endofstartup), .CLK(1'b0), .GSR(1'b0), .GTS(1'b0), .KEYCLEARB(1'b1));

assign reset0 = !endofstartup;
FDPE flipflop1(reset1, clock24, 1'b1, reset0, reset0);
FDPE flipflop2(reset2, clock24, 1'b1, reset1, reset1);
FDPE flipflop3(reset, clock24, 1'b1, reset2, reset2);


// Detecteren of de 27 MHz al draait door de frequentie te vergelijken met
// de 24 MHz-klok op het FPGA-bordje.
// Check if the 27 MHz is already running by comparing the frequency to the
// 24 MHz clock on the FPGA board.

// Drie flip-flops met asynchrone set instantieren!
// Instantiate three flip-flops with asynchronous set!
// FDPE bla (output, clock, clock enable, asynchronous preset, data)
FDPE flipflop4 (resetsynch1, sdclock, 1'b1, reset, reset);
FDPE flipflop5 (resetsynch2, sdclock, 1'b1, resetsynch1, resetsynch1);
FDPE flipflop6 (sdreset1, sdclock, 1'b1, reset, ((!reg8[7])||(reg8[7]&&reg8[5])));
// ((!reg8[7])||(reg8[7]&&reg8[5])) is waar als de frequentie van de 27
// MHz-klok veel te laag of veel te hoog is; ik test ook op te hoog voor het
// geval dat bijvoorbeeld de SN74AHCU04 zelf gaat staan te oscilleren zonder
// kristalsignaal.
// ((!reg8[7])||(reg8[7]&&reg8[5])) is true when the frequency of the 27 MHz
// clock is much too low or much too high; I also check for too high just in
// case that, for example, the SN74AHCU04 would self-oscillate without
// a crystal signal.

FDPE flipflop7 (sdreset2, sdclock, 1'b1, sdreset1, sdreset1);

always @ (posedge sdclock)
begin
	if (resetsynch2) 
		tel3 <= 0;
	else
		tel3 <= (tel3 + 3'b001) % 8;
end
always @ (posedge clock24)
begin
	sdclkdiv8_1 <= tel3[2];
	sdclkdiv8_2 <= sdclkdiv8_1;
	sdclkdiv8_3 <= sdclkdiv8_2;
	if (reset)
	begin
		tel11 <= 0;
		tel8 <= 0;
		reg8 <= 0;
	end
	else
	begin
		if (tel11[10])
		begin
			tel11 <= 0;
			reg8 <= tel8;
			tel8 <= 0;
		end
		else
		begin
			tel11 <= tel11+1;
			if (sdclkdiv8_2 & (~sdclkdiv8_3)) tel8 <= (tel8 + 1) % 256;
		end
	end
end

// Twee a drie seconden extra vertraging maken om de oscillator- en
// klokbufferbuizen goed te laten opwarmen, anders komt er harde ruis
// uit de DAC.
// Make two or three seconds extra delay to let the oscillator and
// clock buffer valves heat up properly, otherwise the DAC produces
// loud noise.
always @(posedge sdclock or posedge sdreset2)
begin
	if (sdreset2)
		delaycount27 <= 27'b0;
	else
	begin
		if (!delaycount27[26]) delaycount27 <= delaycount27 + 27'b1;
	end
end
// FDPE bla (output, clock, clock enable, asynchronous preset, data)
FDPE flipflop7b (sdreset, sdclock, 1'b1, sdreset2, !delaycount27[26]);

// 85A2 aanschakelen en zaadjes produceren
// Switch on 85A2 and produce seeds for the pseudorandom number generators
always @ (posedge clock24)
begin
	if (reset)
	begin
		seedcount1 <= 16'h5e4f; // Zomaar een getal, just picked a number
		seedcount2 <= 16'h7a32; // Zomaar een getal, just picked a number
		switch85A2_int <= 1'b0;
	end
	else
	begin
		switch85A2_int <= 1'b1; // moet uit blijven tijdens configureren!
		// Aangezien HSWAPEN hoog is op de TE0630-module, is dat een
		// kwestie van een weerstandje naar aarde aan de
		// switch85A2-uitgang hangen.
		//
		// 85A2 has to stay off during configuration! As HSWAPEN is
		// made high on the FPGA module, that's a matter of connecting
		// an external pull-down resistor to the switch85A2 output.
		//
		if (~ignited85A2) seedcount1 <= (seedcount1 + 16'b1) % 65536;
		if (sdreset) seedcount2 <= (seedcount2 + 16'b1) % 65536;
		// Hier hebben we een onveilige klokovergang, maar dat kan
		// alleen maar meer toeval opleveren, dus het geeft niet.
		//
		// Unsafe clock transition, but that can only result in more
		// randomness, so it doesn't matter.
	end
end

// Initialiseren DIX4192 en SRC4392
// Initializing DIX4192 and SRC4392
always @ (posedge sdclock)
begin
	if (sdreset)
	begin
		teller <= 15'b000000000000000;
		spirstn_int <= 1'b0;
		spicsn_int <= 1'b1;
		spicclk_int <= 1'b0;
		src4392cdin_int <= 1'b0;
		dix4192cdin_config <= 1'b0;
		updaterreset <= 1'b1;
	end
	else
	begin
		if (~((teller[14])&(teller[10])&(teller[0])))
		       	teller <= teller+1;
	       	else
			updaterreset <= 1'b0;
		if (teller == 24) spirstn_int <= 1'b1;
		if (teller[14])
		begin
			if (~((teller[10])&(teller[0])))
			begin
				spicsn_int <= 1'b0;
				spicclk_int <= teller[0];
				src4392cdin_int <= src4392spidata[~teller[9:1]];
				// Alle bits van teller[9:1] inverteren omdat
				// spidata er van hoog naar laag in moet!
				//
				// Invert all bits of teller[9:1] because
				// spidata has to be shifted in from high to
				// low!
				dix4192cdin_config <= dix4192spidata[~teller[9:1]];
			end
			else
				spicsn_int <= 1'b1;
			// 7 ns houddtijd nodig tussen de laatste dalende
			// flank van de klok en de stijgende
			// van csn. Vandaar de constructie met &(teller[0]).
			//
			// 7 ns of hold time required between the last falling
			// edge of the clock and the rising edge of csn. Hence
			// the construction with &(teller[0]).
		end
		// Je moet 01hex...31hex instellen.
		// Dat komt dus neer op een adres plus 49 keer 8 bits aan
		// data, ruwweg 400 bits. Daar kun je veilig 512 van maken
		// zonder dat er wat overloopt.
		// Elke twee tellen wordt er een bit overgedragen, dus heb je
		// dan 1024 tellen nodig.
		//
		// About 400 bits of data need to be transferred. That can be
		// rounded to 512 bits without any overflows. A bit is
		// transferred at every other value of teller, so you need
		// 1024 teller cycles.
	end
end

// Ervoor zorgen dat de dix4192updater meteen een reset krijgt en pas een paar
// klokslagen na het vrijgeven van updaterreset geen reset meer krijgt.
//
// Ensure that the dix4192updater immediately gets a reset and that reset only gets
// released a few clock cycles after updaterreset.
//
// FDPE bla (output, clock, clock enable, asynchronous preset, data)
FDPE flipflop8(dixupdaterreset1, sdclock, 1'b1, sdreset, updaterreset);
FDPE flipflop9(dixupdaterreset, sdclock, 1'b1, sdreset, dixupdaterreset1);

always @ * knoplampreset_int = dixupdaterreset;

dix4192updater du1(sdclock, dixupdaterreset, sourcesel, rnd, dix4192cdout, dix4192cclk_ud, dix4192cdin_ud, dix4192csn_ud);
assign dix4192csn = spicsn_int && dix4192csn_ud; // beide in rust hoog, both high when in rest.
assign dix4192cclk = spicclk_int || dix4192cclk_ud; // beide in rust laag, both low when in rest.
assign dix4192cdin = dixupdaterreset ? dix4192cdin_config : dix4192cdin_ud; // if dixupdaterreset then config else updater

// Verrassingstoestandsmachine met opslag in I2C EEPROM, surprise state machine with storage in I2C EEPROM
assign writebyte = {surprise[3:0], ~surprise[3:0]};
i2c i2cinterface(sda_in, sda_out, scl, sdclock, sdreset, 16'b0000000000000000, read, readbyte, write, writebyte, i2cready); // Ik denk dat sdreset goed is,
// of moet het toch een latere reset wezen?
// I think sdreset is correct, or must it be a later reset?

// Toestandsmachine voor verrassingsinstelling, inclusief opslaan in EEPROM
// State machine for surprise setting, including storage in EEPROM
always @ (posedge sdclock)
begin
	if (rnd[1:0] != 2'b00) rndd[1:0] <= rnd[1:0]; // rndd[1:0] driewaardig gemaakt, made three-valued
	if (rnd[3:2] != 2'b00) rndd[3:2] <= rnd[3:2]; // rndd[3:2] driewaardig gemaakt, made three-valued
	if (sdreset) // Ik denk dat sdreset goed is. I think sdreset is correct.
	begin
		vt <= 3'b000;
		read <= 1'b0;
		write <= 1'b0;
		i2cready_old <= 1'b1;
	end
	else
	begin
		i2cready_old <= i2cready;
		case (vt)
			3'b000: // begin, geef leesopdracht aan de I2C-interface
				// die de EEPROM aanstuurt
				// start, give read command to I2C interface
				// that controls the EEPROM
			begin
				read <= 1'b1;
				write <= 1'b0;
				vt <= 3'b001;
			end
			3'b001: // wacht tot EEPROM gelezen is
				// wait until EEPROM is read
			begin
				read <= 1'b0;
				write <= 1'b0;
				if (i2cready && (!i2cready_old))
				begin
					if ((readbyte[7:4] == ~readbyte[3:0]) && (readbyte[7:6] != 2'b00) && readbyte[5:4] != 2'b00)
					begin
						vt <= 3'b010;
						surprise <= readbyte[7:4];
					end
					else
						vt <= 3'b011;
				end
				else
					vt <= 3'b001;
			end
			3'b010: // ruststand, idle state
			begin
				read <= 1'b0;
				write <= 1'b0;
				if (newsurpriserequest)
					vt <= 3'b011;
				else
					vt <= 3'b010;
			end
			3'b011: // nieuwe trekking, generate new surprise
			begin
				read <= 1'b0;
				write <= 1'b0;
				surprise <= rndd[3:0];
				vt <= 3'b100;
			end
			3'b100: // nieuwe waarde in EEPROM zetten, MAG NIET
				// VAKER DAN EENS PER 5 ms!
				// write new value to EEPROM, SHOULD NOT BE
				// DONE MORE THAN ONCE PER 5 ms!
			begin
				read <= 1'b0;
				write <= 1'b1;
				vt <= 3'b010;
			end
			default:
			begin
				read <= 1'b0;
				write <= 1'b0;
				vt <= 3'b000;
			end
		endcase
	end
end

// Relaisaansturing, relay control
always @ (posedge sdclock or posedge sdreset)
if (sdreset || knoplampreset)
begin
	relaycounter <= 20'b0;
	waitfori2c <= 1'b1;
end
else
begin
	if (vt == 3'b010) waitfori2c <= 1'b0;
	if ((!waitfori2c) && (!relaycounter[19])) relaycounter <= relaycounter + 20'b1;
end
always @ * relay_int = relaycounter[19] && (!relayoffkl);

assign src4392csn = spicsn_int;
assign src4392cclk = spicclk_int;
assign srcdixrstn = spirstn_int;
assign src4392cdin = src4392cdin_int;
assign switch85A2 = switch85A2_int;
assign knoplampreset = knoplampreset_int;
assign relay = relay_int;
assign seed[31:0] = {seedcount1[0], seedcount1[1], seedcount1[2], seedcount1[3], seedcount2[0], seedcount2[1], seedcount2[2], seedcount2[3], seedcount1[15:4], seedcount2[15:4]};
endmodule
