module sigmadelta(x, m, clock, reset, dither1, dither2, dither3, dither4, rnd1, rnd2, rot, y, qclip, iclip, sclip, scale);
// m: methode (method)
// 0: stilte (silence) 1010101010...
// 1: Risbo-chaos
// 2: PBM4 (PWM4)
// 3: PBM8 (PWM8)
// scale:
// 0: ingang volle schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 0: full-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 1: ingang halve schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 1: half-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 2: ingang kwart schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 2: quarter-scale input drives the sigma-delta to 75 % ones / 25 % ones
// 3: ingang achtste schaal stuurt sigma-delta uit tot 75 % enen / 25 % enen
// 3: eighth-scale input drives the sigma-delta to 75 % ones / 25 % ones
parameter wordlength = 35;
// wordlength: woordlengte van ingangssignaal en van de geherkwantiseerde
// interne signalen
// wordlength: word length of input signal and requantized internal signals
parameter coefflength = 17;
// coefflength: lengte van resonatorterugkoppelcoefficienten en van de
// schaling aan de ingang, geen teken
// coefflength: length of resonator feedback coefficients and of the input
// signal scaling, unsigned
parameter coefflengthmin1 = coefflength - 1;
parameter limitlength = 10;
// limitlength: lengte van begrenzingsniveaus integratoren, geen teken
// limitlength: length of limiting levels of the integrators, unsigned
parameter [limitlength-1:0] max10 = 1;
parameter [limitlength-1:0] max11 = 29;
parameter [limitlength-1:0] max12 = 16;
parameter [limitlength-1:0] max13 = 6;
parameter [limitlength-1:0] max20 = 1;
parameter [limitlength-1:0] max21 = 772;
parameter [limitlength-1:0] max22 = 562;
parameter [limitlength-1:0] max23 = 255;
parameter [limitlength-1:0] max30 = 1;
parameter [limitlength-1:0] max31 = 49;
parameter [limitlength-1:0] max32 = 11;
parameter [limitlength-1:0] max33 = 7;
parameter [limitlength-1:0] max40 = 1;
parameter [limitlength-1:0] max41 = 518;
parameter [limitlength-1:0] max42 = 129;
parameter [limitlength-1:0] max43 = 100;
parameter [limitlength-1:0] max5 = 567;
parameter [limitlength-1:0] max6 = 492;
parameter [coefflength-1:0] h10rec = 0;
parameter [coefflength-1:0] h11rec = 54;
parameter [coefflength-1:0] h12rec = 21;
parameter [coefflength-1:0] h13rec = 5;
parameter [wordlength-2:0] c10 = 0;
parameter [wordlength-2:0] c11 = 14055017;
parameter [wordlength-2:0] c12 = 5451233;
parameter [wordlength-2:0] c13 = 1279589;
parameter [wordlength-2:0] c20 = 0;
parameter [wordlength-2:0] c21 = 722376754;
parameter [wordlength-2:0] c22 = 342499497;
parameter [wordlength-2:0] c23 = 113482549;
parameter [wordlength-2:0] c30 = 0;
parameter [wordlength-2:0] c31 = 66586200;
parameter [wordlength-2:0] c32 = 9640166;
parameter [wordlength-2:0] c33 = 4306764;
parameter [wordlength-2:0] c40 = 0;
parameter [wordlength-2:0] c41 = 974567593;
parameter [wordlength-2:0] c42 = 167570866;
parameter [wordlength-2:0] c43 = 100232789;
// parameter [wordlength-2:0] c5 = 1974121616;
// parameter [wordlength-2:0] c6 = 2602607922;
parameter c5 = 34'd1974121616;
parameter c6 = 34'd2602607922;
parameter [wordlength-2:0] cn0 = 0;
// parameter [wordlength-2:0] cn1 = 2973481568;
parameter cn1 = 34'd2973481568;
parameter [wordlength-2:0] cn2 = 1725873600;
parameter [wordlength-2:0] cn3 = 1348195279;
// cn: c-coefficient van de laatste integrator, dus integrator 5 in een
// PBM-stand of integrator 7 in de Risbochaosstand 
// cn: c-coefficient of last integrator, that is, integrator 5 in a PWM
// mode or integrator 7 in Risbo chaos mode 
parameter [coefflength-1:0] d10 = 0;
parameter [coefflength-1:0] d11 = 1;
parameter [coefflength-1:0] d12 = 1;
parameter [coefflength-1:0] d13 = 1;
parameter [coefflength-1:0] d20 = 0;
parameter [coefflength-1:0] d21 = 10;
parameter [coefflength-1:0] d22 = 10;
parameter [coefflength-1:0] d23 = 10;
parameter [3:0] s210 = 8;  // maakt de mux vast eenvoudiger dan s210 = 0. Probably simplifies mux.
parameter [3:0] s211 = 8;
parameter [3:0] s212 = 10;
parameter [3:0] s213 = 10;
parameter [3:0] s21max = 10;
// s21*: aantal plaatsen dat signaal naar rechts geschoven wordt tussen
// eerste en tweede resonator, het gaat dus maal 2 tot de macht -s21*
// s21*: number of places the signal gets shifted to the right between the
// first and second resonators, that is, it gets multiplied by 2 to the power
// -s21*
input signed [wordlength-1:0] x;
input [1:0] m;
input clock, reset;
input signed [32:0] dither1;
input signed [32:0] dither2;
input signed [32:0] dither3;
input signed [32:0] dither4;
input signed [wordlength-4:0] rnd1;
input signed [wordlength-4:0] rnd2;
input [2:0] rot;
input [1:0] scale;
output y;
output qclip;
output iclip;
output sclip;
reg signed [wordlength+coefflength-1:0] u1;
wire signed [wordlength-1:0] u9, u10, u11, u12, u14, u15, u17, u18, u20, u21;
wire signed [wordlength+coefflength-1:0] u3, u4, u5, u6a;
reg signed [wordlength+coefflength:0] u6b;
wire signed [3:0] u23;
reg signed [wordlength-1:0] u6, u19;
reg signed [wordlength+s21max-1:0] u7;
wire sd;
wire iclip1, iclip2, iclip3, iclip4, iclip5, iclip6, iclip7;
wire sclip1, sclip2, sclip3, sclip4, sclip5;
wire qclip;
reg [limitlength-1:0] max1;
reg [limitlength-1:0] max2;
reg [limitlength-1:0] max3;
reg [limitlength-1:0] max4;
reg [coefflength-1:0] h1rec;
reg [wordlength-2:0] c1;
reg [wordlength-2:0] c2;
reg [wordlength-2:0] c3;
reg [wordlength-2:0] c4;
reg [wordlength-2:0] cn;
reg [coefflength-1:0] d1;
reg [coefflength-1:0] d2;
always @ (*)
begin
	case (m)
		2'b00:
		begin
			max1 = max10;
			max2 = max20;
			max3 = max30;
			max4 = max40;
			case (scale)
				2'b00: h1rec = h10rec;
				2'b01: h1rec = h10rec << 1;
				2'b10: h1rec = h10rec << 2;
				2'b11: h1rec = h10rec << 3;
			endcase
			c1 = c10;
			c2 = c20;
			c3 = c30;
			c4 = c40;
			cn = cn0;
			d1 = d10;
			d2 = d20;
		end
		2'b01:
		begin
			max1 = max11;
			max2 = max21;
			max3 = max31;
			max4 = max41;
			case (scale)
				2'b00: h1rec = h11rec;
				2'b01: h1rec = h11rec << 1;
				2'b10: h1rec = h11rec << 2;
				2'b11: h1rec = h11rec << 3;
			endcase
			c1 = c11;
			c2 = c21;
			c3 = c31;
			c4 = c41;
			cn = cn1;
			d1 = d11;
			d2 = d21;
		end
		2'b10:
		begin
			max1 = max12;
			max2 = max22;
			max3 = max32;
			max4 = max42;
			case (scale)
				2'b00: h1rec = h12rec;
				2'b01: h1rec = h12rec << 1;
				2'b10: h1rec = h12rec << 2;
				2'b11: h1rec = h12rec << 3;
			endcase
			c1 = c12;
			c2 = c22;
			c3 = c32;
			c4 = c42;
			cn = cn2;
			d1 = d12;
			d2 = d22;
		end
		2'b11:
		begin
			max1 = max13;
			max2 = max23;
			max3 = max33;
			max4 = max43;
			case (scale)
				2'b00: h1rec = h13rec;
				2'b01: h1rec = h13rec << 1;
				2'b10: h1rec = h13rec << 2;
				2'b11: h1rec = h13rec << 3;
			endcase
			c1 = c13;
			c2 = c23;
			c3 = c33;
			c4 = c43;
			cn = cn3;
			d1 = d13;
			d2 = d23;
		end
	endcase
end
always @ (posedge clock) u1 <= $signed({1'b0, h1rec})*x;
subc #(.xlength(wordlength+coefflength)) sub1(u1, sd, c1, u6, d1, u3, sclip1);
integrator #(.wordlength(wordlength+coefflength)) int1(u3, clock, reset, u4, iclip1, max1);
subb sub2(u4[wordlength+coefflength-1:coefflength], sd, c2, u5[wordlength+coefflength-1:coefflength], sclip2);
// u3, u4, u5 en u6a zijn extra lang
// u3, u4, u5 and u6a are extra long
assign u5[coefflength-1:0] = u4[coefflength-1:0]; 
// Laagste bits veranderen niet door optellen of aftrekken van c2.
// Lower bits don't change by adding or subtracting c2.
integrator #(.wordlength(wordlength+coefflength)) int2(u5, clock, reset, u6a, iclip2, max2);
always @ *
begin
	u6b = u6a + $signed(dither1[32:32-coefflength]) + $signed({1'b0, 1'b1, {coefflengthmin1{1'b0}}});
	u6[wordlength-1:0] = u6b[wordlength+coefflength-1:coefflength];
	case(m)
		2'b00: u7 = {{wordlength{1'b0}}, {s21max{1'b0}}};
		2'b01: u7 = (u6<<<(s21max-s211));
		2'b10: u7 = (u6<<<(s21max-s212));
		2'b11: u7 = (u6<<<(s21max-s213));
	endcase
end
suba #(.xlength(wordlength+s21max)) sub3(u7, sd, c3, u12, d2, dither2[32:32-coefflength], u9, sclip3);
integrator int3(u9, clock, reset, u10, iclip3, max3);
subb sub4(u10, sd, c4, u11, sclip4);
integrator int4(u11, clock, reset, u12, iclip4, max4);
chaosdc #(.c(c5), .max(max5)) chaos1(u12, sd, dither3[32:27], m, clock, reset, u15, iclip5);
chaosnyquist #(.c(c6), .max(max6)) chaos2(u15, sd, dither4[32:28], m, clock, reset, u18, iclip6);
always @ *
begin
	case(m)
		2'b00: u19 = 0;
		2'b01: u19 = u18;
		2'b10: u19 = u12;
		2'b11: u19 = u12;
	endcase
end
subb sub5(u19, sd, cn, u20, sclip5);
lastintegrator intn(u20, clock, reset, u21, iclip7);
quantizer quant(u21, m, rnd1, rnd2, u23, qclip);
pwm pwm1(u23, m, rot, clock, reset, sd);
assign sclip = sclip1|sclip2|sclip3|sclip4|sclip5;
assign iclip = iclip1|iclip2|iclip3|iclip4|iclip5|iclip6|iclip7;
assign y = sd;
endmodule
