module suba (x, sd, c, fb, d, dither, y, overflow);
// berekent y = x - sd*c - d*fb + dither afgerond op wordlength bits met de
// MSB's uitgelijnd. sd wordt opgevat als +/-1, dus 1 wordt +1 en 0 wordt -1.
// calculates y = x - sd*c - d*fb + dither rounded to wordlength bits with the
// MSB's aligned. sd is interpreted as +/-1, thus 1 becomes +1 and 0 becomes
// -1.
parameter wordlength = 35;
// lengte van fb en y, c is een bit korter want c heeft geen teken.
// length of fb and y, c is one bit shorter because c has no sign.
parameter wordlengthmin1 = wordlength - 1;
parameter coefflength = 17;
// lengte van de resonatorcoefficient d, moet kleiner dan of gelijk aan 
// wordlength zijn.
// length of the resonator coefficient d, must be smaller than or equal to
// wordlength.
parameter coefflengthmin1 = coefflength - 1;
parameter xlength = wordlength;
// lengte van x, moet kleiner dan of gelijk aan wordlength + coefflength zijn
// length of x, must be smaller than or equal to wordlength + coefflength
parameter ditherlength = coefflength + 1;
parameter xshift = wordlength + coefflength - xlength;
input signed [xlength-1:0] x;
input signed [wordlength-1:0] fb;
input sd;
input signed [ditherlength-1:0] dither;
input [wordlength-2:0] c;
input [coefflength-1:0] d;
output signed [wordlength-1:0] y;
output overflow;
reg overflow_int;
reg signed [ditherlength:0] ditherplusoffset;
reg signed [wordlength+coefflength:0] som1;
reg signed [wordlength+coefflength+1:0] som2;
reg signed [wordlength-1:0] y_int;
always @ (*)
begin
	ditherplusoffset = $signed({dither[ditherlength-1], dither[ditherlength-1:0]}) + $signed({1'b0, 1'b1, {coefflengthmin1{1'b0}}});
	if (sd)
		som1 = $signed({x[xlength-1], x[xlength-1:0], {xshift{1'b0}}}) - $signed({1'b0, c, {coefflength{1'b0}}});
	else
		som1 = $signed({x[xlength-1], x[xlength-1:0], {xshift{1'b0}}}) + $signed({1'b0, c, {coefflength{1'b0}}});
	som2 = som1 + ditherplusoffset - fb*$signed({1'b0, d});
	case (som2[wordlength+coefflength+1:wordlength+coefflength-1])
		3'b000:
		begin
			overflow_int = 0;
			y_int[wordlength-1:0] = som2[wordlength+coefflength-1:coefflength];
		end
		3'b111:
		begin
			overflow_int = 0;
			y_int[wordlength-1:0] = som2[wordlength+coefflength-1:coefflength];
		end
		default:
		begin
			overflow_int = 1;
			if (som2[wordlength+coefflength+1])
				y_int[wordlength-1:0] = {1'b1, {wordlengthmin1{1'b0}}};
			else
				y_int[wordlength-1:0] = {1'b0, {wordlengthmin1{1'b1}}};
		end
	endcase
end
assign overflow = overflow_int;
assign y = y_int;
endmodule
