`default_nettype none
module top(src4392csn, src4392cclk, src4392cdin, bck4392, lrck4392, sdml, sdmr,
       dix4192csn, dix4192cclk, srcdixrstn, dix4192cdin, nswitch85A2, scl, sda,
       relay, halfbandlamp, apodizinglamp, steeplamp, chaoslamp, pwm4lamp,
       pwm8lamp, cliplamp, bckout, lrckout, dataout, clock24, sdclock,
       data4392, modeswitch, filtswitch, bron, verrassingsknop, schaalknop,
       qcliplampen, lampenknop, dix4192cdout, ignited85A2, bck, lrck, data,
       deemn, fsgt90kHz_ext, fsgt100kHz_ext, fsgt190kHz_ext);
output wire fsgt90kHz_ext, fsgt100kHz_ext, fsgt190kHz_ext;
input wire clock24, sdclock, data4392;
input wire [3:0] modeswitch;
input wire [3:0] filtswitch;
input wire [3:0] bron;
input wire verrassingsknop;
input wire [2:0] schaalknop;
input wire qcliplampen; // DIL-schakelaartje waarmee je kiest of vastlopen kwantisator tot oplichten oversturingslamp leidt 
// DIP switch with which you choose whether quantizer clipping is shown on clipping lamp
input wire lampenknop; // DIL-schakelaartjes waarmee je kiest of oversturingslamp ook werkt in de verrassingsstanden
// DIP switch with which you choose whether clipping lamp works in surprise modes
input wire dix4192cdout, ignited85A2;
input wire bck, lrck, data; // van DIX4192, from DIX4192
input wire deemn; // ook van DIX4192, also from DIX4192
// Wat moet je met VALID, VBIT en BPERR?
// GPO-instellingen van de DIX4192:
// adres 1B, waarde 06: GPO1 hoog als er niet-audio-data ontvangen worden
// adres 1C, waarde 05: GPO2 laag als er pre-emfase gebruikt is
// adres 1D, waarde 07: GPO3 hoog als er ongeldige data ontvangen worden
// adres 1E, waarde 0D: GPO4 hoog bij een pariteitsfout
// Je hebt ook nog de LOCKN-poot, maar de DIX4192 is zo ingesteld
// dat deze het geluid uitschakelt als de SPDIF-ontvanger ontgrendeld raakt.
//
// Hang GPO1 maar aan de mute-ingang van de SRC4392, gebruik GPO2 om de
// de-emfase te schakelen (deemn = GPO2) en vergeet de rest
// voorlopig maar. Ik weet toch niet wat de timing van het valid-bit en de
// pariteitsfoutdetectie ten opzichte van de I2S is en nul is ook een heel
// fout monster.
//
// What to do with VALID, VBIT and BPERR?
// GPO settings of the DIX4192:
// address 1B, value 06: GPO1 high when non-audio data are received
// address 1C, value 05: GPO2 low when pre-emphasis has been used
// address 1D, value 07: GPO3 high when invalid data are received
// address 1E, value 0D: GPO4 high in case of a parity error
// There is also a LOCKN pin, but the settings of the DIX4192 are such
// that it switches off the sound when the SPDIF receiver looses lock.
//
// Ty GPO1 to the mute input of the SRC4392, use GPO2 to switch
// the de-emphasis (deemn = GPO2) and forget the rest.
// I don't know the timing of the valid bit and the parity error bit
// with respect to the I2S anyway and when you would insert zeroes because
// of invalid data, you would insert a value that is also quite wrong.
//
output wire src4392csn, src4392cclk, src4392cdin, bck4392, lrck4392, sdml, sdmr;
output wire dix4192csn, dix4192cclk, srcdixrstn, dix4192cdin, nswitch85A2, scl, relay;
output wire halfbandlamp, apodizinglamp, steeplamp, chaoslamp, pwm4lamp, pwm8lamp, cliplamp;
output wire bckout, lrckout, dataout; // naar SRC4392, to SRC4392
inout wire sda; 
wire sda_in, sda_out, scl_out;
wire iclipl, qclipl, sclipl, iclipr, qclipr, sclipr, fclipinterpol, fclipprefilter;
assign fclipprefilter = 1'b0; // prefilter.v heeft geen fclip, prefilter.v has no fclip
wire knoplampreset;
wire switch85A2;

wire reset, newsurpriserequest, mod0, dsddet, fsgt90kHz, fsgt100kHz, fsgt190kHz, relayoffkl;
wire [1:0] sourcesel;
wire [5:0] rnd_seq;
wire [3:0] surprise;
wire [1:0] m;
wire [1:0] scale;
wire [2:0] filtsel;
wire anacorr;
wire [31:0] seed;

wire sdml0, sdmr0;
wire signed [63:0] rndlongdither;
reg signed [63:0] rndlongdither_old;
reg sdml1, sdmr1;
// reg sdml2, sdmr2;
reg [1:0] m1;
reg [1:0] m2;
reg [1:0] scale1;
reg [1:0] scale2;
reg anacorr1;
reg anacorr2;
reg deem1, deem2;

reg [2:0] filtsel_synch2;
reg [2:0] filtsel_synch1;
parameter wordlength = 35; // zie sigmadelta.v, as in sigmadelta.v
wire sdreset, firclock, firreset;
wire signed [32:0] ditherfirl1;
wire signed [32:0] ditherfirr1;
wire signed [32:0] ditherfirl2;
wire signed [32:0] ditherfirr2;
wire signed [32:0] dither1l;
wire signed [32:0] dither2l;
wire signed [32:0] dither3l;
wire signed [32:0] dither4l;
wire signed [32:0] dither1r;
wire signed [32:0] dither2r;
wire signed [32:0] dither3r;
wire signed [32:0] dither4r;
reg signed [64:0] longdither; // dither voor prefilter, maar 200 kwoord/s nodig
// dither for prefilter, only 200 kword/s needed

wire signed [wordlength-4:0] rnd1l;
wire signed [wordlength-4:0] rnd1r;
wire signed [wordlength-4:0] rnd2l;
wire signed [wordlength-4:0] rnd2r;
wire [2:0] rotl;
wire [2:0] rotr;
wire [63-wordlength:0] dummy1l;
wire [63-wordlength-6:0] dummy1r; // rnd_seq wordt van het rechterkanaal afgetapt
// rnd_seq is tapped off from the right channel
wire [66-wordlength:0] dummy2l;
wire [66-wordlength:0] dummy2r;

wire signed [wordlength-1:0] interpoloutl;
wire signed [wordlength-1:0] interpoloutr;

wire resetsynch1;
assign mod0 = 0; // mod0 zet sigma-delta in een stand waarbij er alleen
// 10101010... patronen gemaakt worden (stilte dus), maar dit wordt niet gebruikt.
// Bij niet-PCM-data gebruiken we de mute-ingang van de 
// SRC4392 en het validity-bit en pariteitsfouten negeren we, dus is mod0 niet
// nodig.
// mod0 can put the sigma-delta into a mode where it only produces
// 10101010... patterns (silence), but this isn't used.
// For non-PCM data we use the mute input of the SRC4392 and we ignore the
// validity bit and parity errors, so mod0 is not required. 

sequencer sequencer1(clock24, reset, sdclock, sdreset, sourcesel, rnd_seq,
newsurpriserequest, dix4192cdout, src4392csn, dix4192csn, src4392cclk,
dix4192cclk, srcdixrstn, src4392cdin, dix4192cdin, switch85A2, ignited85A2,
seed, scl_out, sda_in, sda_out, relayoffkl, relay, knoplampreset, surprise);

IOBUF #(.DRIVE(6), .IOSTANDARD("LVCMOS33"), .SLEW("SLOW")) iobuf_sda(.O(sda_in), .IO(sda), .I(1'b0), .T(sda_out));
OBUF #(.DRIVE(6), .IOSTANDARD("LVCMOS33"), .SLEW("SLOW")) obuf_scl(.O(scl),
	.I(scl_out)); // Wat verwarrende naam: scl is de feitelijke 
// uitgangspoot van de FPGA terwijl scl_out een intern signaal is, 
// dit om consequent te blijven met de naamgeving bij sda.
// sequencer.v ondersteunt het uitrekken van de klok niet, vandaar OBUF!
//
// Somewhat confusing name: scl is the actual output pin of the FPGA while
// scl_out is an internal signal, just to stay consistent with the naming
// used for sda. sequencer.v does not support clock stretching, hence
// the use of an OBUF!
PULLUP pusda(sda);
PULLUP puscl(scl);

knoplamp knoplamp1(sdclock, knoplampreset, modeswitch, filtswitch, bron,
verrassingsknop, schaalknop, surprise, mod0, dsddet, fsgt90kHz, fsgt100kHz,
fsgt190kHz, qcliplampen, lampenknop, newsurpriserequest, relayoffkl, m,
scale, filtsel, anacorr, sourcesel, halfbandlamp, apodizinglamp, steeplamp,
chaoslamp, pwm4lamp, pwm8lamp, cliplamp, sclipl, sclipr, iclipl, iclipr,
qclipl, qclipr, fclipinterpol, fclipprefilter);

dopdetector dopdetector1(reset, bck, lrck, data, dsddet); // goede reset? correct reset?
fscounter fscounter1(firclock, lrck, firreset, fsgt90kHz, fsgt100kHz, fsgt190kHz);
// De fsgt*kHz-signalen veranderen op de opgaande flanken van de firclock.
// The fsgt*kHz signals change at the rising edges of firclock.
always @ (posedge firclock or posedge reset)
	if (reset)
	begin
		filtsel_synch1 <= 3'b000;
		filtsel_synch2 <= 3'b000;
	end
	else
	begin
		filtsel_synch1 <= filtsel;
		filtsel_synch2 <= filtsel_synch1;
	end
	//
	// filtsel[2], [1] en [0] worden niet noodzakelijkerwijs op dezelfde
	// klokflank ingeklokt als je ze als asynchrone signalen behandelt,
	// maar dat is niet erg; een paar nanoseconden het foute filter erin hebben
	// zal niet opvallen, zolang hij maar niet in een foute toestand komt
	// waar hij nooit meer uit raakt.
	// Misschien zou het mis kunnen gaan bij het interpolerende filter als
	// dat ten onrechte even geselecteerd wordt terwijl de
	// bemonsteringsfrequentie boven de 100 kHz ligt,
	// maar ongebruikt worden alle filters van prefilter.v hoe dan ook
	// gereset.
	//
	//
	// filtsel[2], [1] and [0] are not necessarily clocked in at the same
	// clock edge when you treat them as asynchronous signals,
	// but that should not be a problem; having the wrong filter for a
	// few nanoseconds will not be noticeable, just as long as nothing
	// gets in an incorrect state it doesn't recover from.
	// It might go wrong with the interpolating filter if it is
	// unintendedly selected while the sample rate exceeds 100 kHz,
	// but it will then get reset, as all unused filters in prefilter.v
	// are kept in reset.
	//

prefilter prefilter1(firreset, bck, lrck, data, firclock, filtsel_synch2, longdither, bckout, lrckout, dataout);

interpol interpol1(sdclock, firclock, bck4392, lrck4392, sdreset, data4392, {deem2, anacorr2}, ditherfirl1, ditherfirr1, ditherfirl2, ditherfirr2, interpoloutl, interpoloutr, fclipinterpol);
// De reset van interpol moet goed getimed zijn ten opzichte van firclock, die
// intern van sdclock wordt afgeleid.
// Vanwege de klokvermenigvuldiger in interpol, mag zijn reset echter niet van
// de firclock afhangen!
//
// The reset of interpol must be properly timed with respect to firclock,
// which it internally derives from sdclock. As the clock multiplier is inside
// interpol, its reset may not depend on firclock!
//
sigmadelta sdl(interpoloutl, m2, sdclock, sdreset, dither1l, dither2l, dither3l, dither4l, rnd1l, rnd2l, rotl, sdml0, qclipl, iclipl, sclipl, scale2);
sigmadelta sdr(interpoloutr, m2, sdclock, sdreset, dither1r, dither2r, dither3r, dither4r, rnd1r, rnd2r, rotr, sdmr0, qclipr, iclipr, sclipr, scale2);

dithergen dgfir1(firclock, firreset, {seed, 32'h03156D10}, ditherfirl1, ditherfirr1);
dithergen dgfir2(firclock, firreset, {seed[31:24], 32'hCD125F23, seed[23:0]}, ditherfirl2, ditherfirr2);

dithergen dg1(sdclock, sdreset, {32'h09F4BE77, seed[31:0]}, dither1l, dither1r);
dithergen dg2(sdclock, sdreset, {16'h26FE, seed[31:0], 16'h9803}, dither2l, dither2r);
dithergen dg3(sdclock, sdreset, {seed[31:8], 32'hCD567200, seed[7:0]}, dither3l, dither3r);
dithergen dg4(sdclock, sdreset, {32'h0025D8C2, seed}, dither4l, dither4r);
xorshiftplus xspl1(sdclock, sdreset, {16'h4583, seed[31:16], 16'h2AB7, seed[15:0]}, {rnd1l, rotl, dummy1l});
xorshiftplus xspr1(sdclock, sdreset, {seed[7:0], 8'h59, seed[15:8], 8'h81, seed[23:16], 8'h25, seed[31:24], 8'hEE}, {rnd1r, rotr, rnd_seq, dummy1r});
xorshiftplus xspl2(sdclock, sdreset, {32'h93205FF3, seed[7:0], seed[31:8]}, {rnd2l, dummy2l});
xorshiftplus xspr2(sdclock, sdreset, {seed[31:13], 32'hF2863025, seed[12:0]}, {rnd2r, dummy2r});
xorshiftplus xsppf(sdclock, sdreset, {seed[0], 16'hF286, seed[3:1], 16'h5FF5, seed[31:4]}, rndlongdither);
always @ (posedge sdclock)
begin
	rndlongdither_old <= rndlongdither;
	longdither <= $signed(rndlongdither) + $signed(rndlongdither_old);
end

always @ (posedge sdclock)
begin
	sdml1 <= sdml0;  //Pijpleidingerij, pipe lining
	sdmr1 <= sdmr0;
	if (sdreset)
	begin
		m1 <= 2'b0;
		m2 <= 2'b0;
		scale1 <= 2'b0;
		scale2 <= 2'b0;
	end
	else
	begin
		m1 <= m;
		m2 <= m1; //Synchroniseren, synchronizing
		scale1 <= scale;
		scale2 <= scale1; //Synchroniseren, synchronizing
	end
end
assign sdml = sdml1;
assign sdmr = sdmr1;

always @ (posedge firclock)
begin
	if (firreset)
	begin
		anacorr1 <= 1'b1;
		anacorr2 <= 1'b1; // 1 betekent dat hij wel corrigeert, 1 means that it does correct
		deem1 <= 1'b0;
		deem2 <= 1'b0; // 0 betekent geen deemfase, 0 means no deemphasis
	end
	else
	begin
		anacorr1 <= anacorr;
		anacorr2 <= anacorr1; //Synchroniseren, synchronizing
		deem1 <= ~deemn;
		deem2 <= deem1; //Synchroniseren, synchronizing
	end
end

// Resetsynchronisatiemeuk?
// reset en sdreset worden in sequencer gemaakt.
// Dan heb je firreset nog over.
// firreset moet je hier synchroniseren met de firclock.
// Twee flip-flops met asynchrone set instantieren.
//
// Reset synchronization stuff.
// reset and sdreset are generated inside sequencer.
// That leaves firreset.
// firreset gets synchronized to the firclock here,
// by means of two instantiated flip-flops with asynchronous set.
FDPE flipflop1 (resetsynch1, firclock, 1'b1, sdreset, sdreset);
FDPE flipflop2 (firreset, firclock, 1'b1, resetsynch1, resetsynch1);

assign fsgt90kHz_ext = fsgt90kHz;
assign fsgt100kHz_ext = fsgt100kHz;
assign fsgt190kHz_ext = fsgt190kHz;
assign nswitch85A2 = ~switch85A2;

endmodule
`default_nettype wire
